#include <signal.h>
#include <stdlib.h>

#include <iostream>
#include <sstream>
#include <string>

using namespace std;

#include "6510.h"
#include "debugger.hpp"
#include "stilinfo.hpp"

MOS6510 *cpu;

const char CBMchar[128] = {
  /* 00 */ '@','A','B','C','D','E','F','G',
  /* 08 */ 'H','I','J','K','L','M','N','O',
  /* 10 */ 'P','Q','R','S','T','U','V','W',
  /* 18 */ 'X','Y','Z','[','',']','^','.',
  /* 20 */ ' ','!','"','#','$','%','&','\'',
  /* 28 */ '(',')','*','+',',','-','.','/',
  /* 30 */ '0','1','2','3','4','5','6','7',
  /* 38 */ '8','9',':',';','<','=','>','?',
  /* 40 */ 'x','x','x','x','x','x','x','x',
  /* 48 */ 'x','x','x','x','x','x','x','x',
  /* 50 */ 'x','x','x','x','x','x','x','x',
  /* 58 */ 'x','x','x','x','x','x','x','x',
  /* 60 */ 'x','x','x','x','x','x','x','x',
  /* 68 */ 'x','x','x','x','x','x','x','x',
  /* 70 */ 'x','x','x','x','x','x','x','x',
  /* 78 */ 'x','x','x','x','x','x','x','x',
};

const char PETchar[128] = {
  /* 00 */ '@','a','b','c','d','e','f','g',
  /* 08 */ 'h','i','j','k','l','m','n','o',
  /* 10 */ 'p','q','r','s','t','u','v','w',
  /* 18 */ 'x','y','z','[','',']','^','.',
  /* 20 */ ' ','!','"','#','$','%','&','\'',
  /* 28 */ '(',')','*','+',',','-','.','/',
  /* 30 */ '0','1','2','3','4','5','6','7',
  /* 38 */ '8','9',':',';','<','=','>','?',
  /* 40 */ '.','A','B','C','D','E','F','G',
  /* 48 */ 'H','I','J','K','L','M','N','O',
  /* 50 */ 'P','Q','R','S','T','U','V','W',
  /* 58 */ 'X','Y','Z','.','.','.','.','.',
  /* 60 */ ' ','.','.','.','.','.','.','.',
  /* 68 */ '.','.','.','.','.','.','.','.',
  /* 70 */ '.','.','.','.','.','.','.','.',
  /* 78 */ '.','.','.','.','.','.','.','.',
};

void registers()
{
  cpu->info();
}

void pcaxy(istream& i, int& pc, int& a, int& x, int& y)
{
  string s;
  getline(i, s);
  istringstream is(s);

  is >> s;
  if(is)
    {
      pc=strtol(s.c_str(), 0, 16);
      is >> s;
      if(is)
	{
	  a=strtol(s.c_str(), 0, 16);
	  is >> s;
	  if(is)
	    {
	      x=strtol(s.c_str(), 0, 16);
	      is >> s;
	      if(is)
		{
		  y=strtol(s.c_str(), 0, 16);
		}
	    }
	}
    }
}

void stopcpu(int s)
{
  cpu->abort=true;
}

void debugger(C64 *c64)
{
  if(c64==0)
    return;

  sighandler_t int_h=signal(SIGINT, stopcpu);
  sighandler_t term_h=signal(SIGTERM, stopcpu);

  cpu=c64->cpu();
  cpu->debug=true;
  registers();

  bool run=true;
  string command;
  while(run)
    {
      string line; getline(cin, line);
      istringstream is(line);
      if(line.size())
	{
	  is >> command;
	}

      if(command=="q" || command=="x")
	run=false;
      else if(command=="Q")
	exit(0);
      // assemble
      else if(command=="a")
	{
	}
      // toggle breakpoint
      else if(command=="b")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  cpu->breakpointToggle(pc);
	  registers();
	}
      // print breakpoints
      else if(command=="bp")
	{
	  cpu->breakpointPrint();
	  registers();
	}
      // clear breakpoints
      else if(command=="bc")
	{
	  cpu->breakpointClear();
	  registers();
	}
      // compare
      else if(command=="c")
	{
	}
      // disassemble
      else if(command=="d")
	{
	  static int pcs=-1;
	  int pc=-1, a=-1, x=-1, y=-1;
	  if(pc==-1)
	    pc=pcs;
	  pcaxy(is, pc, a, x, y);
	  pcs=cpu->disass(pc);
	}
      // fill
      else if(command=="f")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  if(x==-1)
	    x=1;
	  if(pc<0 || pc>0xFFFF)
	    {
	      printf("addr out of range: %04X\n", pc);
	      goto nofill;
	    }
	  if(pc+x-1>0xFFFF)
	    {
	      printf("size exceeds 0xFFFF\n");
	      goto nofill;
	    }
	  if(a<0 || a>0xFF)
	    {
	      printf("value of of range: %X\n", a);
	      goto nofill;
	    }
	  memset(c64->memory()+pc, a, x);
	  printf("filled %X bytes\n", x);
	nofill: ;
	}
      // go
      else if(command=="g")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  infostr.clear();
	  int cyc=cpu->jsr(pc, a, x, y);
	  cout << "executed " << cyc << " cycles" << endl;
	  for(infostr_t::iterator i=infostr.begin(); i!=infostr.end(); ++i)
	    cout << *i << endl;

	  registers();
	}
      // single step
      else if(command==".")
	{
	  int pc=-1, a=0, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc>=0 && pc<=0xFFFF)
	    cpu->set(pc,a,x,y);
	  infostr.clear();
	  int cyc=cpu->parse();
	  cout << "executed " << cyc << " cycles" << endl;
	  for(infostr_t::iterator i=infostr.begin(); i!=infostr.end(); ++i)
	    cout << *i << endl;
	  registers();
	}
      else if(command=="pc")
	{
	  int pc=-1, a=0, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc>=0 && pc<=0xFFFF)
	    cpu->set(pc,a,x,y);
	  registers();
	}
      // hunt
      else if(command=="h")
	{
	}
      // load
      else if(command=="l")
	{
	  int addr;
	  FILE *f;
	  size_t size;
	  string fn, addrs;
	  is >> fn;
	  if(fn.empty())
	    {
	      cerr << "no filename" << endl;
	      goto noload;
	    }
	  is >> addrs;
	  addr=strtol(addrs.c_str(), 0, 16);
	  if(addr<0 || addr>0xFFFF)
	    {
	      cerr << "load address out of range: " << addr << endl;
	      goto noload;
	    }

	  f=fopen(fn.c_str(), "rb");
	  if(f==0)
	    {
	      perror("could not open file");
	      goto noload;
	    }

	  if(fseek(f, 0, SEEK_END) < 0)
	    {
	      perror("could not seek");
	      goto abortload;
	    }

	  size=ftell(f);
	  rewind(f);

	  if(addr+size-1>0xFFFF)
	    {
	      cerr << "size of file would exceed 0xFFFF: " << size << endl;
	      goto abortload;
	    }

	  if(fread(c64->memory()+addr, size, 1, f)!=1)
	    {
	      perror("could not read entire file");
	    }

	  printf("loaded $%04X bytes\n", size);

	abortload:
	  fclose(f);
	noload: ;
	}
      // memory display
      else if(command=="m")
	{
	  int pc=-1, a=-1, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(a==-1)
	    a=0x10;
	  if(pc>=0 && pc<=0xFFFF)
	    {
	      printf("%04X: ", pc);
	      for(; pc<=0xFFFF && a; pc++, a--)
		printf("%02X ", c64->getmem(pc));
	      printf("\n");
	    }
	}
      // registers
      else if(command=="r")
	{
	  registers();
	}
      // save
      else if(command=="s")
	{
	}
      // copy/transport
      else if(command=="t")
	{
	}
      // toggle memory watch
      else if(command=="w")
	{
	  int pc=-1, a=-1, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc<0 || pc>0xFFFF)
	    printf("invalid address %x\n", pc);
	  else
	    c64->watchToggle(pc);
	}
      // clear watchpoints
      else if(command=="wc")
	{
	  c64->watchClear();
	  printf("cleared watchpoints\n");
	}
      // print watchpoints
      else if(command=="wp")
	{
	  c64->watchPrint();
	}
      // help
      else if(command=="?")
	{
	}
      // reset cpu
      else if(command=="reset")
	{
	  cpu->reset();
	  registers();
	}
      else if(command=="dec2hex")
	{
	  string s; is >> s;
	  int a=atoi(s.c_str());
	  printf("%X\n", a);
	}
      else if(command=="hex2dec")
	{
	  string s; is >> s;
	  int a=strtol(s.c_str(), 0, 16);
	  printf("%i\n", a);
	}
      // print current text screen
      else if(command=="screen")
	{
	  string s;
	  is >> s;
	  const char *t=(s=="p")?PETchar:CBMchar;

	  const int bank=(c64->getmem(0xDD00)^3)&3;
	  const int base=c64->getmem(0xD018)>>4;
	  const int offset=bank*0x4000+base*0x0400;
	  printf("screen at %04X\n", offset);
	  byte *p=c64->memory()+offset;
	  for(int y=0; y<25; y++)
	    {
	      for(int x=0; x<40; x++)
		printf("%c", t[(*p++)&0x7F]);
	      printf("\n");
	    }
	}
    }

  cpu->debug=false;
  signal(SIGINT, int_h);
  signal(SIGTERM, term_h);
}
