/*
  sid (c) 1999-2003 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /home/cvs/sidblaster/sid/playlist.cc,v 1.8 2003/05/22 12:18:16 doj Exp $
*/

#include <curses.h>
#undef clear

#include <dirent.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <algorithm>

using namespace std;

#include "playlist.h"
#include "stilinfo.hpp"

extern char *errstr;

Playlist::Playlist() :
  current_(0),
  cursor(0),
  playw(0),
  width(0),
  height(0)
{ }

Playlist::~Playlist()
{
  if(playw)
    delwin(playw);
}

void Playlist::init(int& xoffset, int& yoffset, int height_)
{
  if(playw)
    delwin(playw), playw=0;

  width=COLS-xoffset;
  if(width<1)
    return;
  height=height_;
  if(height<1)
    return;
  playw=newwin(height, width, yoffset, xoffset);
  if(playw==0)
    {
      errstr="could not create playw";
      exit(1);
    }

  xoffset=COLS;
  yoffset+=height;
}

void Playlist::playinfo_print()
{
  if(!playw)
    return;

  wclear(playw);

  const int s=size();
  int i=cursor-height/2;
  if(i<0)
    i=0;
  if(i>s-height)
    i=s-height;
  if(i<0)
    i=0;
  int y=0;
  while(i<s && y<height)
    {
      if(i==cursor)
	wattron(playw, A_REVERSE);
      if(i==current_)
	wattron(playw, A_BOLD);
      mvwaddnstr(playw, y++, 0, a[i].c_str(), width);
      if(i==cursor)
	wattroff(playw, A_REVERSE);
      if(i==current_)
	wattroff(playw, A_BOLD);

      i++;
    }

  if(wnoutrefresh(playw)!=OK)
    {
      errstr="error on refresh of trackw";
      exit(1);
    }
}

bool isDir(const string& filename)
{
  struct stat filestat; // structure for a file status
  // if the status can not be fetched we skip processing of this file
  if(lstat(filename.c_str(), &filestat) == -1)
    return false;

  if(S_ISDIR(filestat.st_mode))
    return true;

  return false;
}

vector<string> readDir(const string& dirname)
{
  vector<string> v;
  if(!isDir(dirname))
    goto readDirError;

  DIR *dir; // structure to store a directory

  // open the directory
  if((dir=opendir(dirname.c_str())) == 0) 
    goto readDirError;

  // parse all entries in the directory
  struct dirent *file; // structure for a directory entry
  while((file=readdir(dir)) != 0) 
    {
      // the files '.' and '..' only cause trouble. We skip them
      if( (strcmp(file->d_name, ".") == 0) ||
	  (strcmp(file->d_name, "..") == 0) )
	continue;
      
      v.push_back(dirname+"/"+file->d_name);
    }

  closedir(dir);

 readDirError:
  return v;
}

Playlist& Playlist::add(const string& s, const bool sortit, const bool addDir)
{
  const string fn=absFilename(s);

  if(isDir(fn))
    {
      vector<string> v=readDir(fn);
      for(vector<string>::const_iterator i=v.begin(); i!=v.end(); ++i)
	add(*i, false, true);
    }
  else
    a.push_back(fn);

  if(sortit && !addDir)
    sort();

  playinfo_print();
  return *this;
}

Playlist& Playlist::next()
{
  if(current_<static_cast<int>(size())-1)
    {
      current_++;
      cursor=current_;
      playinfo_print();
    }
  return *this;
}

Playlist& Playlist::previous()
{
  if(current_>0)
    {
      current_--;
      cursor=current_;
      playinfo_print();
    }
  return *this;
}

Playlist& Playlist::random()
{
  current_=static_cast<int>(static_cast<double>(rand())/static_cast<double>(RAND_MAX)*size());
  cursor=current_;
  playinfo_print();
  return *this;
}

const string Playlist::current()
{
  if(current_<static_cast<int>(size()))
    return a[current_];
  return "";
}

Playlist& Playlist::randomize()
{
  const int e=rand()%10;
  for(int i=0; i<e; i++)
    random_shuffle(a.begin(), a.end());

  playinfo_print();
  return *this;
}

Playlist& Playlist::sort()
{
  std::sort(a.begin(), a.end());
  playinfo_print();
  return *this;
}

void Playlist::up()
{
  if(cursor>0)
    {
      cursor--;
      playinfo_print();
    }
}

void Playlist::down()
{
  if(cursor<static_cast<int>(size())-1)
    {
      cursor++;
      playinfo_print();
    }
}
