/*
  sid (c) 1999-2003 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /home/cvs/sidblaster/sid/stilinfo.cpp,v 1.5 2003/05/22 01:20:56 doj Exp $
*/

#include "stil.h"
#include <string>
#include <vector>
#include <unistd.h>

using namespace std;

#include <curses.h>
#undef clear

extern char *errstr;
extern char musinfostring[10][33];

#include "stilinfo.hpp"

infostr_t infostr;

std::string absFilename(std::string fn)
{
  // if relative path?
  if(fn[0]!='/')
    {
      char buf[PATH_MAX];
      getcwd(buf, PATH_MAX);
      fn=string(buf)+"/"+fn;
    }

  // remove leading /
  fn.erase(0, 1);

  // condense . and .. in path
  typedef vector<string> v_t;
  v_t v;
  do {
    // ./
    if(fn[0]=='.' && fn[1]=='/')
      {
	fn.erase(0, 2);
      }
    // ../
    else if(fn[0]=='.' && fn[1]=='.' && fn[2]=='/')
      {
	if(v.size())
	  v.pop_back();
	fn.erase(0, 3);
      }
    // a dir/filename
    else
      {
	const size_t p=fn.find('/');
	if(p==string::npos)
	  {
	    v.push_back(fn);
	    fn.erase();
	  }
	else
	  {
	    v.push_back(string(fn, 0, p));
	    fn.erase(0, p+1);
	  }
      }
    
  } while(fn.size());
  
  fn='/';
  for(v_t::iterator i=v.begin(); i!=v.end();)
    {
      fn+=*i;
      ++i;
      if(i!=v.end())
	fn+='/';
    }

  return fn;
}

static WINDOW *stilw=0;
static bool initialized=false;
static int infostry=0;
static int infostrlastsize=-1;
static int stilwWidth;

static void cleanup()
{
  if(stilw)
    delwin(stilw), stilw=0;
}

void stilinfo_init(int& xoffset, int& yoffset)
{
  if(!initialized)
    {
      atexit(cleanup);
      initialized=true;
    }

  // delete windows
  cleanup();

  stilwWidth=COLS-xoffset;
  stilw=newwin(LINES-yoffset, stilwWidth, yoffset, xoffset);
  if(!stilw)
    {
      errstr="could not init stilw";
      exit(1);
    }
  infostry=LINES-yoffset-1;

  xoffset=COLS;
  yoffset=LINES;
}

void stilinfo_print(STIL *stil, const std::string& fn, int tuneno)
{
  wclear(stilw);
  if(stil && fn.size())
    {
      string f=absFilename(fn);
      waddstr(stilw, stil->getAbsGlobalComment(f.c_str()));
      waddstr(stilw, stil->getAbsEntry(f.c_str(), tuneno));
      waddstr(stilw, stil->getAbsBug(f.c_str(), tuneno));
    }

  for(int i=0; i<10; i++)
    mvwaddstr(stilw, i, 0, musinfostring[i]);

  wnoutrefresh(stilw);

  infostrlastsize=-1;
}

void stilinfo_printinfostr()
{
  const size_t s=infostr.size();
  if(s && static_cast<int>(s)!=infostrlastsize)
    {
      infostrlastsize=s;
      wattron(stilw, A_REVERSE);
      int y=infostry;
      for(infostr_t::iterator i=infostr.begin(); i!=infostr.end() && y>=0; ++i, --y)
	{
	  mvwaddstr(stilw, y, 0, *i);
	  int l=stilwWidth-strlen(*i)-1;
	  while(l-->0)
	    waddch(stilw, ' ');
	}
      wattroff(stilw, A_REVERSE);
      wnoutrefresh(stilw);
    }
}
