#include <curses.h>
#include <stdint.h>
#include <stdlib.h>

#include <cwsid.h>

#include "c64.h"
#include "sidinfo.h"

extern char *errstr;

static WINDOW* trackw[CWSID_MAX_NUM];
static bool initialized;
static int scrollspeed=4;

static void cleanup()
{
  for(int i=0; i<CWSID_MAX_NUM; i++)
    if(trackw[i])
      {
	delwin(trackw[i]);
	trackw[i]=0;
      }
}

void trackinfo_init(int& xoffset, int& yoffset, C64 *c64)
{
  if(!initialized)
    {
      atexit(cleanup);
      initialized=true;
    }

  cleanup();

  for(int i=0; i<c64->sidNum(); i++)
    {
      trackw[i]=newwin(LINES-yoffset, 12, yoffset, xoffset);
      if(trackw[i]==0)
	{
	  errstr="could not create trackw. maybe terminal width too small?";
	  exit(1);
	}
      scrollok(trackw[i], TRUE);
      for(int l=0; l<LINES+1; l++)
	waddstr(trackw[i], "\n... ... ...");

      xoffset+=12;
    }
  yoffset=LINES;
}

// we exploit the fact, that those strings a and b are only 3(4) chars long
inline bool noteequal(char *a, char *b)
{
  uint32_t *aa=reinterpret_cast<uint32_t*>(a);
  uint32_t *bb=reinterpret_cast<uint32_t*>(b);
  return *aa==*bb;
}

// same exploit as in noteequal()
inline void notecopy(char *dst, char *src)
{
  uint32_t *d=reinterpret_cast<uint32_t*>(dst);
  uint32_t *s=reinterpret_cast<uint32_t*>(src);
  *d=*s;
}

void trackinfo_print(C64 *c64)
{
  static int c=0;
  if((++c)%scrollspeed)
    return;

  if(c64==0)
    return;

  allSID allsid;
  c64->get(allsid);
  const bool pal=c64->getPal();

  static word freq[CWSID_MAX_NUM*3];
  static char freqstr[4][CWSID_MAX_NUM*3];

  for(int i=0; i<c64->sidNum(); i++)
    {
      SID &sid=allsid.sid[i];
      WINDOW *w=trackw[i];

      waddch(w, '\n');

      for(int j=0; j<3; j++)
	{
	  const size_t voicenum=i*3+j;

	  if(c64->voiceActive(voicenum+1))
	    {
	      SIDvoice &v=sid.voice[j];
	      const word f=lohi(v.freqLo, v.freqHi);
	      char *fstr=freq2note(static_cast<double>(f) * (pal?985248.0:1.0227E6) / 16777216.0);

	      // check if a new note has been triggered
	      if(c64->gateTriggered(voicenum))
		{
		  wattron(w, A_BOLD);
		  waddstr(w, fstr);
		  wattroff(w, A_BOLD);
		}
	      else
		{
		  // check if a note note has been slided into
		  if(!noteequal(fstr, freqstr[voicenum]))
		    {
		      wattron(w, A_DIM);
		      waddstr(w, fstr);
		      wattroff(w, A_DIM);
		    }
		  else
		    {
		      // first col
		      if(v.test)
			waddch(w, 't');
		      else
			waddch(w, '.');

		      // second col
		      if(v.ring)
			waddch(w, 'r');
		      else if(f!=freq[voicenum])
			waddch(w, '~');
		      else
			waddch(w, '.');
		      
		      // third col
		      if(v.sync)
			waddch(w, 's');
		      else
			waddch(w, '.');
		    }
		}
	      
	      freq[voicenum]=f;
	      notecopy(freqstr[voicenum], fstr);
	    }
	  else // voiceActive
	    waddstr(w, "...");

	  if(j<2)
	    waddch(w, ' ');
	}

      if(wnoutrefresh(trackw[i])!=OK)
	{
	  errstr="error on refresh of trackw";
	  exit(1);
	}
    }

  c64->setGateBits(false);
}

void trackinfo_scrollspeed(int s)
{
  if(s>0)
    scrollspeed=s;
}
